local BufferTable = {}
function BufferReached(key, buffer)
  if key == nil then return end
  if BufferTable[key] == nil then BufferTable[key] = {} end
  BufferTable[key].buffer = buffer or 3
  BufferTable[key].now = GetFrameTimeSeconds()
  if BufferTable[key].last == nil then BufferTable[key].last = BufferTable[key].now end
  BufferTable[key].diff = BufferTable[key].now - BufferTable[key].last
  BufferTable[key].eval = BufferTable[key].diff >= BufferTable[key].buffer
  if BufferTable[key].eval then BufferTable[key].last = BufferTable[key].now end
  return BufferTable[key].eval
end

local A = nil   -- Instance of InventoryMod

InventoryMod = {}
InventoryMod.__index = InventoryMod
InventoryMod.fn = {}

InventoryMod._author = "Cr4x"
InventoryMod._name = "InventoryMod"
InventoryMod._version = {
	major = 1,
	api = 08,
	build = 1,
}
InventoryMod._slash = "/im"

InventoryMod.defaultVars =
  {
    activeFilter =
    {
      [INVENTORY_BACKPACK] = { },
      [INVENTORY_BANK] = { },
      [INVENTORY_GUILD_BANK] = { },
    },
    filterTab =
    {
      -- All filter
      [0] =
      {
        1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,
        50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,
      },
      -- Weapon filter
      [1] =
      {
        IM_FILTER_WEAPON_1H_AXE, IM_FILTER_WEAPON_1H_HAMMER, IM_FILTER_WEAPON_1H_SWORD, IM_FILTER_WEAPON_2H_SWORD, IM_FILTER_WEAPON_2H_AXE, IM_FILTER_WEAPON_2H_HAMMER,
        IM_FILTER_WEAPON_BOW, IM_FILTER_WEAPON_STAFF, IM_FILTER_WEAPON_DAGGER
      },
      -- Armor filter
      [2] =
      {
        ITEMTYPE_COSTUME, ITEMTYPE_DISGUISE, ITEMTYPE_TABARD, IM_FILTER_ARMOR_HEAVY, IM_FILTER_ARMOR_MEDIUM, IM_FILTER_ARMOR_LIGHT, IM_FILTER_WEAPON_SHIELD
      },
      -- Consumables filter
      [3] =
      {
        ITEMTYPE_DRINK, ITEMTYPE_FOOD, ITEMTYPE_POTION, ITEMTYPE_RECIPE, ITEMTYPE_POISON, ITEMTYPE_AVA_REPAIR
      },
      -- Materials filter
      [4] =
      {
        ITEMTYPE_BLACKSMITHING_BOOSTER, ITEMTYPE_BLACKSMITHING_MATERIAL, ITEMTYPE_BLACKSMITHING_RAW_MATERIAL, IM_FILTER_BLACKSMITH,
        ITEMTYPE_CLOTHIER_BOOSTER, ITEMTYPE_CLOTHIER_MATERIAL, ITEMTYPE_CLOTHIER_RAW_MATERIAL, IM_FILTER_CLOTHIER,
        ITEMTYPE_ENCHANTING_RUNE, ITEMTYPE_ENCHANTMENT_BOOSTER, IM_FILTER_ENCHANT,
        ITEMTYPE_WOODWORKING_BOOSTER, ITEMTYPE_WOODWORKING_MATERIAL, ITEMTYPE_WOODWORKING_RAW_MATERIAL, IM_FILTER_WOODWORKER,
        ITEMTYPE_ALCHEMY_BASE, ITEMTYPE_REAGENT, IM_FILTER_ALCHEMIST,
        ITEMTYPE_FLAVORING, ITEMTYPE_INGREDIENT, ITEMTYPE_RAW_MATERIAL, ITEMTYPE_SPICE, ITEMTYPE_STYLE_MATERIAL,
        ITEMTYPE_WEAPON_TRAIT, ITEMTYPE_ARMOR_TRAIT, ITEMTYPE_ARMOR_BOOSTER
      },
      -- Others filter
      [5] =
      {
        ITEMTYPE_ADDITIVE, ITEMTYPE_COLLECTIBLE, ITEMTYPE_CONTAINER,
        ITEMTYPE_GLYPH_ARMOR, ITEMTYPE_GLYPH_JEWELRY, ITEMTYPE_GLYPH_WEAPON,
        ITEMTYPE_LOCKPICK, ITEMTYPE_LURE, ITEMTYPE_PLUG, ITEMTYPE_SCROLL,
        ITEMTYPE_SOUL_GEM, ITEMTYPE_TOOL, ITEMTYPE_TROPHY, ITEMTYPE_SIEGE, ITEMTYPE_TRASH
      },
    },
	
	filter = 
	{
	
	},
  }

setmetatable(InventoryMod, {
  __call = function(class, ...)
    local self = setmetatable({}, InventoryMod)
    self.inventoryAccount = {} -- SavedVars: all account inventories [BAG_BANK] | [BAG_GUILDBANK]
    self.inventoryCharacter = {} -- SavedVars: all character inventories [BAG_BACKPACK]
    self.vars = {}   -- SavedVars: all vars
    self.handlers = {}
    
    self.guildBankActive = 0     -- 0 Inactive else the guild id
    
    --self.settings:Initialize()
    return self
  end,
})

local BAG_INVENTORYTYPE =
  {
    [BAG_BACKPACK] = INVENTORY_BACKPACK,
    [BAG_BANK] = INVENTORY_BANK,
    [BAG_GUILDBANK] = INVENTORY_GUILD_BANK
  }

local INVENTORYTYPE_BAG =
  {
    [INVENTORY_BACKPACK] = BAG_BACKPACK,
    [INVENTORY_BANK] = BAG_BANK,
    [INVENTORY_GUILD_BANK] = BAG_GUILDBANK
  }


--[[
REGION: functions
--]]

local function fnCopy (t) -- shallow-copy a table
  if type(t) ~= "table" then return t end
  local meta = getmetatable(t)
  local target = {}
  for k, v in pairs(t) do target[k] = v end
  setmetatable(target, meta)
  return target
end

local function ApplyFilter(inventoryType)
  inventoryType = inventoryType or INVENTORY_BACKPACK
  local esoFilterId = PLAYER_INVENTORY.inventories[inventoryType].currentFilter

  local filterId = A.vars.activeFilter[inventoryType][esoFilterId]

  if ( filterId ~= nil and A.vars.filter[filterId] ~= nil and A.vars.filter[filterId].data[inventoryType] ~= nil ) then
    local inventory = PLAYER_INVENTORY.inventories[inventoryType]
    local list = inventory.listView
    if (list) then
      local scrollData = ZO_ScrollList_GetDataList(list)
      ZO_ScrollList_Clear(list)

      local slots = inventory.slots
      for slotIndex, slotData in pairs(slots) do
        if(PLAYER_INVENTORY:ShouldAddSlotToList(inventory, slotData)) then
          local slotId = slotData.slotIndex
          if ( A.vars.filter[filterId].data[inventoryType][slotIndex] ) then
            scrollData[#scrollData + 1] = ZO_ScrollList_CreateDataEntry(inventory.listDataType, slotData)
          end
        end
      end
    end

    ZO_ScrollList_Commit(list)
  end
end

local function GetCleanString(stringId)
  if (stringId == nil or stringId == "" ) then do return "" end end
  local name = GetString(stringId)
  local pos = zo_strfind(name, "^", 0, 1)
  if ( pos ~= nil ) then
    name = zo_strsub(name, 0, pos-1)
  end
  return name
end

local function GetFilterName(k, v)
	local name = ""
	if ( type(v) == "number" ) then
		name = GetCleanString(v)
	elseif ( type(v) == "string" ) then
		local id = ""
		if ( v == "SI_ITEMTYPE" ) then
			id = zo_loadstring("return "..v..k)
		elseif ( v == "" or v == " " ) then
			name = name..v
		else
			id = zo_loadstring("return "..v)
		end
		
		if ( id == nil ) then
			name = name..v
		elseif not ( id == "" ) then
			name = GetCleanString(id())
		end
	elseif ( type(v) == "table" ) then
		for _, v2 in pairs(v) do
			name = name..GetFilterName(k, v2)
		end
	end
	
	if ( zo_strlen(name) ) then
		do return name end
	end
end

local function IsMatchingFilter(slotData, obj, fAnd)
  fAnd = fAnd or false

  if ( type(obj) == "function" and obj(slotData) ) then
    do return true end
  elseif ( type(obj) == "number" and IsMatchingFilter(slotData, A.vars.filter[obj].filter) ) then
    do return true end
  elseif ( type(obj) == "string" ) then
    local tmp = zo_loadstring(obj)
    if ( type(tmp) == "function" and IsMatchingFilter(slotData, tmp() ) ) then
      do return true end
    end
  elseif ( type(obj) == "userdata" or type(obj) == "table" ) then
    for k, v in pairs(obj) do
      local fMatched = IsMatchingFilter(slotData, v)
      if ( fAnd and not fMatched ) then
        do return false end
      elseif ( not fAnd and fMatched ) then
        do return true end
      end
    end

    if ( fAnd ) then
      do return true end
    end
  end

  return false
end

local function VersionToString()
	return InventoryMod._version.major.."."..InventoryMod._version.api.."."..InventoryMod._version.build
end

local function PostHook(object, funcName, callback)
  local tmp = object[funcName]
  if (tmp) then
    object[funcName] = function(...)
      tmp(...)
      callback(...)
    end
  end
end

--[=[
local function SaveInventory(_bagId)
  do return end
  local slots = nil
  if ( _bagId == nil or _bagId == BAG_BACKPACK ) then
    slots = copy(PLAYER_INVENTORY.inventories[BAG_INVENTORYTYPE[_bagId]].slots)
    if ( slots ~= nil ) then
      for k,v in pairs(slots) do
        slots[k].dataEntry = nil
        slots[k].inventory = nil
      end
      A.inventoryCharacter = slots
    end
  else
    slots = copy(PLAYER_INVENTORY.inventories[BAG_INVENTORYTYPE[_bagId]].slots)
    if ( slots ~= nil ) then
      for k,v in pairs(slots) do
        slots[k].dataEntry = nil
        slots[k].inventory = nil
      end
      A.inventoryAccount = slots
    end
  end
end
]=]--

local function SetHandler(self, event, callback, handler)
  if ( self == nil ) then do return end end
  local name = self:GetName()
  local ZO_handler = self:GetHandler(event)

  if ( A.handlers[name] == nil ) then
    A.handlers[name] = {}
    A.handlers[name][event] = {}
    if ( ZO_handler ~= nil ) then
      table.insert(A.handlers[name][event], ZO_handler)
    end
  elseif ( A.handlers[name][event] == nil ) then
    A.handlers[name][event] = {}
    if ( ZO_handler ~= nil ) then
      table.insert(A.handlers[name][event], ZO_handler)
    end
  end

  table.insert(A.handlers[name][event], callback)
  self:SetHandler(event, handler)
end

--[[
ENDREGION: functions
--]]

--[[
REGION: events
--]]

local function OnCloseGuildBank(eventCode)
  A.guildBankActive = 0
end

local function OnGlobalMouseUp(eventCode, button, ctrl, alt, _shift, command)
  if ( button == 1 or button == 2 ) then
    local obj = WINDOW_MANAGER:GetMouseOverControl()
    if ( obj ~= nil and obj:GetOwningWindow() == InventoryModControl and not A.Menu:IsHidden()) then
      A.Menu:SetHidden(true)
    end
  end
end

local function OnGuildBankSelected(eventCode, guildId)
  A.guildBankActive = guildId
end

local function OnGuildBankItemsReady(eventCode)
  A:RefreshFilter(INVENTORY_GUILD_BANK)
  --SaveInventory(BAG_GUILDBANK)
end

local function OnHideInventory(self, hidden)
  if ( not A.Menu:IsHidden() ) then
    A.Menu:SetHidden(true)
  end
end

local function OnHideHandler(self, hidden)
  local name = self:GetName()
  for _,v in pairs(A.handlers[name]["OnHide"]) do
    v(self, hidden)
  end
  return false
end

local function OnItemSingleSlotUpdate(eventCode, bagId, slotId, isNewItem, itemSoundCategory, updateReason)
  if ( bagId == BAG_BACKPACK or bagId == BAG_BANK or bagId == BAG_GUILDBANK ) then
    A:RefreshFilterData(BAG_INVENTORYTYPE[bagId])
  end
end

local function OnMouseUpFilterTab(self, button, isInside, ctrl, alt, _shift, command)
  if ( button == 1 ) then
    if ( not A.Menu:IsHidden() ) then
      A.Menu:SetHidden(true)
    end
  elseif ( button == 2 ) then
    
    local esoFilter = self.m_object.m_buttonData.filterType
    local inventoryType = self.m_object.m_buttonData.inventoryType
    local data = {}
    
    for _, filterId in pairs(A.vars.filterTab[esoFilter]) do
      if ( A.vars.filter[filterId] ~= nil and A.vars.filter[filterId].data[inventoryType] ~= nil and next(A.vars.filter[filterId].data[inventoryType]) ) then
        data[#data + 1] = {
          text = A.vars.filter[filterId].name,
          filterId = filterId
        }
      end
    end

    if ( #data > 0 ) then
      table.sort(data, function(e1, e2)
        return e1.text < e2.text
      end)

      local all = GetString(IM_SI_ALL)

      table.insert(data, 1, {
        text = all
      })

      A.Menu.Controls.scrollControl.dataEntry.data = data

      A.Menu:ClearAnchors()
      A.Menu:SetAnchor(TOPLEFT, self, BOTTOM)
      A.Menu:SetRow(0)
      A.Menu:SetHidden(false)
    end
  end
end

local function OnMouseUpFilterTabHandler(self, button, isInside, ctrl, alt, _shift, command)
  local name = self:GetName()
  for _,v in pairs(A.handlers[name]["OnMouseUp"]) do
    v(self, button, isInside, ctrl, alt, _shift, command)
  end
  return false
end

function InventoryMod.OnMouseUpMenuItem(self, button, isInside, ctrl, alt, _shift, command)
  local filterId = self.dataEntry.data.filterId

  if ( button == 1 and isInside ) then
    local _, _, tab = A.Menu:GetAnchor()
    local tabOwner = tab:GetOwningWindow()

    local esoFilterId = tab.m_object.m_buttonData.filterType
    local inventoryType = tab.m_object.m_buttonData.inventoryType

    A.vars.activeFilter[inventoryType][esoFilterId] = filterId

    if ( PLAYER_INVENTORY.inventories[inventoryType].currentFilter == esoFilterId ) then
      PLAYER_INVENTORY:UpdateList(inventoryType)
    end
    A.Menu:SetHidden(true)

  end
end

local function OnMouseUpGuildSearchBox(self, button, isInside, ctrl, alt, _shift, command)
  if ( button == 2 and isInside ) then
    self:SetText("")
  end
end

local function OnMouseUpSearchBox(self, button, isInside, ctrl, alt, _shift, command)
  if ( button == 2 and isInside ) then
    self:SetText("")
  end
end

local function OnMouseUpSearchBoxHandler(self, button, isInside, ctrl, alt, _shift, command)
  local name = self:GetName()
  for _,v in pairs(A.handlers[name]["OnMouseUp"]) do
    v(self, button, isInside, ctrl, alt, _shift, command)
  end
  return false
end

local function OnShowBank(self, hidden)
  if ( not hidden ) then
    A:RefreshFilterData(INVENTORY_BANK)
    ZO_PlayerBankSearchBox:SetHidden(false)
    
    PLAYER_INVENTORY:UpdateList(INVENTORY_BANK)
  end
end

local function OnShowGuildBank(self, hidden)
  if ( not hidden ) then
    PLAYER_INVENTORY:UpdateList(INVENTORY_GUILD_BANK)
  -- Searchbox of guild bank is currently not implemented correctly by ZO so hide it
  --ZO_GuildBankSearchBox:SetHidden(false)
  end
end

local function OnShowInventory(self, hidden)
  if ( not hidden ) then
    local searchBox = ZO_PlayerInventorySearchBox
    local menuBarLabelInventory = ZO_PlayerInventoryMenuBarLabel
    local menuBarLabelBank = ZO_PlayerBankMenuBarLabel

    searchBox:ClearAnchors()
    if ( A.bankActive ) then
      searchBox:SetAnchor(RIGHT, menuBarLabelBank, LEFT, -20, 0)
    else
      searchBox:SetAnchor(RIGHT, menuBarLabelInventory, LEFT, -20, 0)
    end
    searchBox:SetHidden(false)
  end
end

local function OnShowHandler(self, hidden)
  local name = self:GetName()
  for _,v in pairs(A.handlers[name]["OnShow"]) do
    v(self, hidden)
  end
  return false
end

--[[
ENREGION: events
--]]

function InventoryMod.fn.FilterByItemtype(slotData, _itemType)
  local itemType = GetItemType(slotData.bagId, slotData.slotIndex)

  return itemType == _itemType
end

function InventoryMod.OnGetDropdown(self)
  local dropdown = InventoryModSettingsFilterDropdownDropdown
  dropdown.m_comboBox:ClearItems()
  local set = {}
  for k,v in pairs(A.vars.filter) do
     set[k] = v.name
  end
  table.sort(set, function(a, b) return a < b end)
  for k,v in pairs(set) do
    dropdown.m_comboBox:AddItem(v)
  end
end

function InventoryMod:ResetFilterData(inventoryType)
  inventoryType = inventoryType or INVENTORY_BACKPACK

  for filterId, filter in pairs(self.vars.filter) do
    if ( self.vars.filter[filterId].data == nil ) then
      self.vars.filter[filterId].data = {}
    end
    self.vars.filter[filterId].data[inventoryType] = {}
  end
end

function InventoryMod:RefreshFilterData(inventoryType)

  inventoryType = inventoryType or INVENTORY_BACKPACK

  local inventory = PLAYER_INVENTORY.inventories[inventoryType]
  local slots = inventory.slots

  self:ResetFilterData(inventoryType)

  local set = {}

  for filterId, obj in pairs(self.vars.filter) do
    for k,v in pairs(slots) do
      if ( IsMatchingFilter(v, obj.filter, true) ) then
        self.vars.filter[filterId].data[inventoryType][v.slotIndex] = true
      end
    end
  end

end

function InventoryMod:RefreshFilter(inventoryType)
  inventoryType = inventoryType or INVENTORY_BACKPACK
  self:RefreshFilterData(inventoryType)
end

function InventoryMod:RefreshAllFilter()
  self:RefreshFilter(INVENTORY_BACKPACK)
  self:RefreshFilter(INVENTORY_BANK)
  self:RefreshFilter(INVENTORY_GUILD_BANK)
end

local function InitializeControls()

  local searchBoxBank = ZO_PlayerBankSearchBox
  local menuBarLabelBank = ZO_PlayerBankMenuBarLabel
  searchBoxBank:ClearAnchors()
  searchBoxBank:SetAnchor(RIGHT, menuBarLabelBank, LEFT, -20, 0)
  searchBoxBank:SetDimensions(240, 24)

  --local searchBoxGuildBank = ZO_GuildBankSearchBox
  --local menuBarLabelGuildBank = ZO_GuildBankMenuBarLabel
  --searchBoxGuildBank:ClearAnchors()
  --searchBoxGuildBank:SetAnchor(RIGHT, menuBarLabelGuildBank, LEFT, -20, 0)
  --searchBoxGuildBank:SetDimensions(240, 24)
end

local function InitializeEvents()
  local tabPlayerInventory = ZO_PlayerInventoryTabs
  local tabBank = ZO_PlayerBankTabs
  local tabGuildBank = ZO_GuildBankTabs

  -- All default tabs in the inventory, -1 for the existing LABEL which just display the current active
  local numTabsInventory = tabPlayerInventory:GetNumChildren() - 1
  local numTabsBank = tabBank:GetNumChildren() - 1
  local numTabsGuildBank = tabGuildBank:GetNumChildren() - 1

  for tabId = 1, numTabsInventory, 1 do
    local tab = tabPlayerInventory:GetNamedChild("Button"..tabId)
    SetHandler(tab, "OnMouseUp", OnMouseUpFilterTab, OnMouseUpFilterTabHandler)
  end

  for tabId = 1, numTabsBank, 1 do
    local tab = tabBank:GetNamedChild("Button"..tabId)
    SetHandler(tab, "OnMouseUp", OnMouseUpFilterTab, OnMouseUpFilterTabHandler)
  end

  for tabId = 1, numTabsGuildBank, 1 do
    local tab = tabGuildBank:GetNamedChild("Button"..tabId)
    SetHandler(tab, "OnMouseUp", OnMouseUpFilterTab, OnMouseUpFilterTabHandler)
  end

  SetHandler(ZO_GuildBank, "OnShow", OnShowGuildBank, OnShowHandler)
  SetHandler(ZO_GuildBank, "OnHide", OnHideInventory, OnHideHandler)
  SetHandler(ZO_PlayerBank, "OnShow", OnShowBank, OnShowHandler)
  SetHandler(ZO_PlayerBank, "OnHide", OnHideInventory, OnHideHandler)
  SetHandler(ZO_GuildBankSearchBox, "OnMouseUp", OnMouseUpGuildSearchBox, OnMouseUpSearchBoxHandler)
  SetHandler(ZO_PlayerBankSearchBox, "OnMouseUp", OnMouseUpSearchBox, OnMouseUpSearchBoxHandler)
  SetHandler(ZO_PlayerInventory, "OnShow", OnShowInventory, OnShowHandler)
  SetHandler(ZO_PlayerInventory, "OnHide", OnHideInventory, OnHideHandler)
  SetHandler(ZO_PlayerInventorySearchBox, "OnMouseUp", OnMouseUpSearchBox, OnMouseUpSearchBoxHandler)

  InventoryModControl:RegisterForEvent(EVENT_GUILD_BANK_SELECTED, OnGuildBankSelected)
  InventoryModControl:RegisterForEvent(EVENT_GUILD_BANK_ITEMS_READY, OnGuildBankItemsReady)
  InventoryModControl:RegisterForEvent(EVENT_CLOSE_GUILD_BANK, OnCloseGuildBank)
  InventoryModControl:RegisterForEvent(EVENT_INVENTORY_SINGLE_SLOT_UPDATE, OnItemSingleSlotUpdate)
  InventoryModControl:RegisterForEvent(EVENT_GLOBAL_MOUSE_UP, OnGlobalMouseUp)

  CALLBACK_MANAGER:RegisterCallback(InventoryMod._name.."MenuOnMouseUp", A.OnMouseUpMenuItem)
  CALLBACK_MANAGER:RegisterCallback("InventoryModSettingsOnGetDropdown", A.OnGetDropdown)

  ZO_PreHook(PLAYER_INVENTORY, "ApplySort", function(self, inventoryType)
    if ( inventoryType == INVENTORY_BACKPACK or inventoryType == INVENTORY_BANK or inventoryType == INVENTORY_GUILD_BANK ) then
      ApplyFilter(inventoryType)
    end
  end)
end

local function InitializeLocale()
  for k, v in pairs(InventoryMod.si) do
    ZO_CreateStringId(k, v)
  end
end

local function InitializeLocaleOnVars()
	if not ( A.defaultVars.filter == nil ) then
		for k,v in pairs(A.defaultVars.filter) do
			A.vars.filter[k].name = GetFilterName(k,v.name);
		end
	end
end

local function OnSlashCommand(args)
  d("InventoryMod", "Version: "..VersionToString(), "Author: "..InventoryMod._author)
end

local function OnAddOnLoaded(eventCode, addOnName)
  if (addOnName ~= InventoryMod._name) then
    return
  end

  A = InventoryMod()

  
  InitializeLocale()

  InitializeEvents()
  InitializeControls()

  A.Menu = ScrollWindow(InventoryMod._name.."Menu", InventoryModControl, nil)
  A.Menu:SetRowDimensions(175,25)

  A.vars = ZO_SavedVars:New( "InventoryModSavedVariables" , 1 , "vars" , InventoryMod.defaultVars )
  
  InitializeLocaleOnVars()
  
  A.inventoryAccount = ZO_SavedVars:NewAccountWide( "InventoryModSavedVariables" , 1 , "inventory" , A.inventoryAccount )
  A.inventoryCharacter = ZO_SavedVars:New( "InventoryModSavedVariables" , 1 , "inventory" , A.inventoryCharacter )

  A:RefreshFilter(INVENTORY_BACKPACK)
  --SaveInventory(BAG_BACKPACK)

  SLASH_COMMANDS[InventoryMod._slash] = OnSlashCommand
end

EVENT_MANAGER:RegisterForEvent(InventoryMod._name.."_EVENT_ADD_ON_LOADED", EVENT_ADD_ON_LOADED, OnAddOnLoaded)