local BACKDROP_CENTER = {
	file = [[/esoui/art/miscellaneous/dropdown_center.dds]],
	size = nil,
	addressMode = nil,
}

local BACKDROP_EDGE = {
	file = [[/esoui/art/miscellaneous/dropdown_edge.dds]],
	width = 128,
	height = 16,
	size = 5,
	padding = 1,
}

local SLIDER = {
	textureFile = [[/esoui/art/miscellaneous/scrollbox_elevator.dds]],
	disabledFile = nil,
	highlightedFile = nil,
	thumbWidth = 16,
	thumbHeight = 32,
	top = nil,
	left = nil,
	bottom = nil,
	right = nil,
}

ScrollWindow.Controls = {}
ScrollWindow.Controls.__index = ScrollWindow

function ScrollWindow.Controls:Initialize(name, parent, contextId, childCount)
	self.childCount = childCount or 8
	self.currentRow = 0
	self.mainPadding =
	{
		x = 10,
		y = 10,
	}
	self.rowHeight = 25
	self.rowPadding = 0
	self.rowWidth = 200
	self.scrollDimensions =
	{
		x = 0,
		y = 0
	}
	self.scrollPadding =
	{
		x = 0,
		y = 0,
	}
	contextId = contextId or ""
	
	local ScrollBarOnValueChanged = function(self, value, reason)
		
	end
	self.name = name..contextId
	self.mainControl = assert(WINDOW_MANAGER:CreateControl(self.name, parent, CT_CONTROL), "Failed to create main CONTROL")
	local mainControl = self.mainControl
	if ( mainControl ~= nil ) then
		-- data
		mainControl.dataEntry = self
		
		-- options
		mainControl:SetClampedToScreen(true)
		mainControl:SetMouseEnabled(true)
		mainControl:SetHidden(true)
		mainControl:SetResizeToFitPadding(10, 10)
	else
		do return false end
	end
	
	self.backDropControl = assert(WINDOW_MANAGER:CreateControl("$(parent)BG", self.mainControl, CT_BACKDROP), "Failed to create background BACKDROP")
	local backDropControl = self.backDropControl
	if ( backDropControl ~= nil ) then
		backDropControl:SetAnchorFill(mainControl)
		backDropControl:SetCenterColor(0,0,0,0.8)
		backDropControl:SetEdgeTexture(BACKDROP_EDGE.file, 
			BACKDROP_EDGE.width,
			BACKDROP_EDGE.height,
			BACKDROP_EDGE.size,
			BACKDROP_EDGE.padding)
		--backDropControl:SetCenterTexture(BACKDROP_CENTER.file,
		--	BACKDROP_CENTER.size,
		--	BACKDROP_CENTER.addressMode)
	end
	
	self.highlightControl = WINDOW_MANAGER:CreateControl("$(parent)Highlight", self.mainControl, CT_BACKDROP)
	local highlightControl = self.highlightControl
	if ( highlightControl ~= nil ) then
		highlightControl:SetHidden(true)
	else
		assert("Failed to create highlight BACKDROP")
	end
	
	self.scrollControl = assert(WINDOW_MANAGER:CreateControl("$(parent)List", self.mainControl, CT_SCROLL), "Failed to create scroll Control")
	local scrollControl = self.scrollControl
	if ( scrollControl ~= nil ) then
		-- data
		scrollControl.dataEntry = {}
		scrollControl.dataEntry.data = {}
		
		-- options
		scrollControl:SetAnchor(LEFT, self.mainControl, LEFT, zo_floor(self.mainPadding.x/2), zo_floor(self.mainPadding.y/2))
		scrollControl:SetMouseEnabled(true)
		
		-- events
		scrollControl:SetHandler("OnMouseWheel", function(self, delta, ctrl, alt, _shift, command)
			local dataCount = #scrollControl.dataEntry.data
			local _self = self:GetParent().dataEntry

			if ( delta < 0 and dataCount > _self.childCount and _self.currentRow < (dataCount - _self.childCount) ) then
				_self.currentRow = _self.currentRow + 1
				_self:RefreshScrollList()
			elseif ( delta > 0 and _self.currentRow > 0 ) then
				_self.currentRow = _self.currentRow - 1
				_self:RefreshScrollList()
			end
		end)
		
	end
	
	self.sliderControl = WINDOW_MANAGER:CreateControl("$(parent)ScrollBar", self.mainControl, CT_SLIDER)
	local sliderControl = self.sliderControl
	if ( sliderControl ~= nil ) then
		-- options
		sliderControl:SetAnchor(RIGHT, self.mainControl, RIGHT, 0, 0)
		sliderControl:SetDimensions(16, 32)
		sliderControl:SetHidden(true)
		sliderControl:SetThumbTexture(
			SLIDER.textureFile,
			SLIDER.disabledFile,
			SLIDER.highlightedFile,
			SLIDER.thumbWidth,
			SLIDER.thumbHeight,
			SLIDER.top, SLIDER.left, SLIDER.bottom, SLIDER.right)
		
		-- events
		sliderControl:SetHandler("OnValueChanged", ScrollBarOnValueChanged)
	else
		do return false end
	end
	
	self.childControls = {}
	self.childControlsHandlers = {}
	for childId = 1, self.childCount, 1 do
		self.childControls[childId] = WINDOW_MANAGER:CreateControl("$(parent)Row"..childId, self.scrollControl, CT_CONTROL)
		local child = self.childControls[childId]
		
		if ( child ~= nil ) then
			--data
			child.dataEntry = {}
			child.dataEntry.data = {}
			child.dataEntry.handlers = {}
			
			--positioning
			child:ClearAnchors()
			if ( childId == 1 ) then
				child:SetAnchor(TOP, self.scrollControl, TOP, zo_floor(self.scrollPadding.x/2), zo_floor(self.scrollPadding.y/2))
			else
				child:SetAnchor(TOP, self.scrollControl, BOTTOM, 0, self.rowPadding)
			end
			--options
			child:SetMouseEnabled(true)
			--events
			local _self = self
			child:SetHandler("OnMouseUp", function(self, button, isInside, ctrl, alt, _shift, command)
				CALLBACK_MANAGER:FireCallbacks(_self.name.."OnMouseUp", self, button, isInside, ctrl, alt, _shift, command)
			end)
			child:SetHandler("OnMouseEnter", function(self)
				if ( highlightControl ~= nil ) then
					highlightControl:ClearAnchors()
					highlightControl:SetAnchorFill(self)
					highlightControl:SetHidden(false)
				end
				CALLBACK_MANAGER:FireCallbacks(_self.name.."OnMouseEnter", self)
			end)
			
			child:SetHandler("OnMouseExit", function(self)
				if ( highlightControl ~= nil ) then
					highlightControl:ClearAnchors()
					highlightControl:SetAnchorFill(menuControl)
					highlightControl:SetHidden(true)
				end
				CALLBACK_MANAGER:FireCallbacks(_self.name.."OnMouseLeave", self)
			end)
			
			local label = WINDOW_MANAGER:CreateControl("$(parent)Label", child, CT_LABEL)
			if ( label ~= nil ) then
				label:SetAnchorFill(child)
				label:SetFont("ZoFontGameSmall")
				label:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
				label:SetText(nil)
				label:SetVerticalAlignment(TEXT_ALIGN_CENTER)
			end
		end
	end
	
	--debug purposes
	self.mainControl:SetAnchor(CENTER, GuiRoot, CENTER)
	return true
end

function ScrollWindow.Controls:Hide()
	self.mainControl:SetHidden(true)
end

function ScrollWindow.Controls:RefreshScrollList()
	local childCount = self.childCount
	local currentRow = self.currentRow
	local data = self.scrollControl.dataEntry.data
	local lastRow = nil
	local scrollHeight = 0
	
	for rowId = 1, childCount, 1 do
		if ( data[currentRow + rowId] ~= nil ) then
			-- data
			self.childControls[rowId].dataEntry.data = data[currentRow + rowId]
			-- positioning
			if ( lastRow == nil ) then
				self.childControls[rowId]:SetAnchor(TOP, self.scrollControl, TOP, zo_floor(self.scrollPadding.x/2), zo_floor(self.scrollPadding.y/2))
			else
				self.childControls[rowId]:SetAnchor(TOP, lastRow, BOTTOM, 0, self.rowPadding)
			end
			lastRow = self.childControls[rowId]
			
			-- label
			local label = self.childControls[rowId]:GetNamedChild("Label")
			if ( label ~= nil ) then
				local text = ""
				if ( type(data[currentRow + rowId].text) == "string" ) then
					text = data[currentRow + rowId].text
				elseif( type(data[currentRow + rowId]) == "string" ) then
					text = data[currentRow + rowId].text
				end
				label:SetText(text)
			end
			
			-- options
			self.childControls[rowId]:SetDimensions(self.rowWidth, self.rowHeight)
			self.childControls[rowId]:SetHidden(false)
			
			scrollHeight = scrollHeight + self.rowHeight + self.rowPadding
		else
			self.childControls[rowId].dataEntry.data = {}
			self.childControls[rowId]:SetHidden(true)
		end
	end
	if ( #data > childCount - 1 ) then
		scrollHeight = scrollHeight - self.rowHeight - self.rowPadding
	end

	self.scrollDimensions.x = self.rowWidth + self.scrollPadding.x
	self.scrollDimensions.y = scrollHeight + self.scrollPadding.y
	self.scrollControl:SetDimensions(self.scrollDimensions.x, self.scrollDimensions.y)
end

function ScrollWindow.Controls:SetRowDimensions(width, height)
	local childCount = 0
	if ( type(width) == "number" and type(height) == "number" ) then
		self.rowWidth = width
		self.rowHeight = height
	end
	if ( not self.mainControl:IsHidden() ) then
		self:RefreshScrollList()
	end
end

function ScrollWindow.Controls:Show()
	self:RefreshScrollList()
	self.mainControl:SetDimensions(self.scrollDimensions.x + self.mainPadding.x, self.scrollDimensions.y + self.mainPadding.y)
	self.mainControl:SetHidden(false)
end